/*
m_rtmidi_freq - Convert MIDI note to frequency in Hz with tuning

DESCRIPTION
This opcode is for MIDI triggered instruments. It converts the active MIDI
note to a frequency in Hz, including pitchbend information, coarse tune
in semitones, fine tune in cents, an optional frequency ratio (useful for audio
rate modulation) and an optional microtuning scale.

SYNTAX
kFreq m_rtmidi_freq kCoarse, kFine [, kRatio, kTuning, kBendDown, kBendUp]

INITIALIZATION

PERFORMANCE
kFreq - Resultant output frequency
kCoarse - Coarse tuning in semitones, where 0 is no change
kFine - Fine tune in cents, 0-127
kRatio - Frequency multiplier, internally quantised to whole ratios, i.e.
	3, 2, 1, 1/2, 1/3, ..., defaults to 1
kTuning - Number of a function table containing a microtuning scale using
	GEN51, defaults to -1, which internally uses equal tempered 12 tone scale
kBendDown, kBendUp - Pitchbend ranges in semitones (down and up respectively)
	default: -1, which means 2 semitones up and down

CREDITS
Author: Jeanette C.
*/

opcode m_rtmidi_freq, k, kkPJJJ
	kCoarse, kFine, kRatio, kTuning, kBendDown, kBendUp xin

	; Initialise variables
	kDownFac init 2 ; pitchbend down scaling
	kUpFac init 2 ; pitchbend up scaling
	kFreq init 0 ; output frequency
	kFineCents = (kFine - 63) / 100
	kRealNote init 0 ; note value with coarse tune offset

	; Default equal tempered 12-tone scale
	iEqual12 = ftgenonce(0, 0, -144, -51, 12, 2, 220, 57, \
		1, 2^(1/12), 2^(2/12), 2^(3/12), 2^(4/12), 2^(5/12), 2^(6/12), \
		2^(7/12), 2^(8/12), 2^(9/12), 2^(10/12), 2^(11/12))
	if (i(kTuning) == -1) then
		kScale init iEqual12
		kScaleLen init 143
		kMinFreq init tab_i(0, iEqual12)
		kMaxFreq init tab_i(143,iEqual12)
	else
		kScale = kTuning
		kScaleLen = tableng(kTuning)
		kMinFreq = tablekt(0,kTuning)
		kMaxFreq = tablekt(kScaleLen,kTuning)
	endif

	; Set effective ratio parameter
	if (kRatio >= 1) then
		kRealRatio = round(kRatio)
	else
		kRealRatio = 1/round(1/kRatio)
	endif
	
	; Set pitchbend up and down ranges
	if (kBendDown == -1) then
		kDownFac = 2
	else
		kDownFac = kBendDown
	endif

	if (kBendUp == -1) then
		kUpFac = 2
	else
		kUpFac = kBendUp
	endif

	; Set tuning scale
	if (changed(kTuning) != 0) then
		if (kTuning == -1) then
			kScale = iEqual12
		else
			kScale = kTuning
		endif
		kScaleLen = tableng(kScale) - 1
		kMinFreq = tablekt(0,kScale)
		kMaxFreq = tablekt(kScaleLen,kScale)
	endif

	; Get MIDI input values
	iNote = notnum()
	kMidiBend = pchbend:k(0,1)

	; Rescale pitchbend
	if (kMidiBend <= 0) then
		kBend = kMidiBend * kDownFac
	else
		kBend = kMidiBend * kUpFac
	endif

	; Calculate the real note
	kRealNote = limit:k((iNote + kCoarse + kBend + kFineCents),0,kScaleLen)

	; Calculate the output frequency
	kFreq = tableikt(kRealNote, kScale)
	kFreq = limit:k((kFreq * kRealRatio), kMinFreq, kMaxFreq)

	; Output the frequency
	xout(kFreq)
endop
